; LCD.ASM
; General purpose LCD display routines.

; Pin Connections:

;       Port B,  Bit 0 = output, RS line to LCD Display
;                    1 = output, EN line to LCD
;                    2 = spare
;                    3 = spare
;                    4 = output, LSB data to LCD
;                    5 = output,     data to LCD
;                    6 = output,     data to LCD
;                    7 = output, MSB data to LCD

; Variables used:
;               LCDtemp
;               High_Count
;               Low_Count

;-------------------------------------------------------
; Initialise the LCD Display.

; Set up for 4 bit data transfers and then send Data in two nibbles
; (each contained in D4 - D7). Send High nibble first.
; Keep D0 & D2 low (these are control lines)
                                                         
InitLCD         movlw   .62             ; 15 msec delay
                call    Delay

; data length is 8 at this point

                movlw   B'00110000'     ; "set data length to 8"
                call    Strobe          ; eg. this will reset it to a known state

                movlw   B'00110000'     ; do that a number of times to make SURE
                call    Strobe

                movlw   B'00110000'     ; again
                call    Strobe

                movlw   B'00100000'     ; "set data length to four"
                call    Strobe          ; (it MUST be ready to go now)

;      o/k data length now 4 bits

                movlw   B'00100000'     ; "set data length to four"
                call    Strobe          ; output nibble one
                movlw   B'10000000'
                call    Strobe          ; output nibble two

                movlw   B'00000000'     ; "Mode set, display shift, inc address"
                call    Strobe          ; output nibble one
                movlw   B'01100000'
                call    Strobe          ; output nibble two

                movlw   B'00000000'     ; "Set display ON, cursor OFF, blink OFF"
                call    Strobe          ; output nibble one
                movlw   B'11000000'     
                call    Strobe          ; output nibble two

                movlw   B'00000000'     ; "Reset and blank display"
                call    Strobe          ; output nibble one
                movlw   B'00010000'
                call    Strobe          ; output nibble two
                return

; -------------------------------------------------------------------
; Display decimal digits in MSD MID & LSD.
; (enter at No_Zero to suppress leading zeros)

No_Zeros        movf    MSD,W           ; get data
                btfss   STATUS,Z        ; but skip any leading zeros
                goto    Dec001
                movf    MID,W
                btfss   STATUS,Z        ; skip another zero
                goto    Dec002
                goto    Dec003

Dec_Out         movf    MSD,W
Dec001          call    Ascii
                movf    MID,W
Dec002          call    Ascii
Dec003          movf    LSD,W
                call    Ascii
                return
                              
; -------------------------------------------------------------------
; Output a character to the LCD.
; The character to be displayed should be in W reg.
; Enter at "Ascii" to convert character to Ascii if necessary
; or at "Display" if not.

Ascii           addlw   .48             ; convert to Ascii
Display         movwf   LCDtemp         ; save the character
                call    WriteData       ; send high nibble
                swapf   LCDtemp,W       ; move low nibble up to high nibble
                call    WriteData
                return

;-------------------------------------------------------
; Initialise starting point on the LCD
;
;       W reg holds location on display
;               Line one reads 000h to 00Fh  (0 to 15) left to right
;               Line two reads 040h to 04Fh  (64 to 79)
;

Position        movwf   LCDtemp         ; save data
                andlw   B'01110000'     ; set control lines to zero, and
                xorlw   B'10000000'     ; set bit 7 high for "Write to Address"
                call    Strobe

                swapf   LCDtemp,W       ; move low nibble up to high nibble
                andlw   B'11110000'     ; strip of low byte
                call    Strobe          ; send low nibble
                return

;-------------------------------------------------------
; Pulse the "Enable" line to clock out a Character to LCD.
;       Enter at WriteData for "Write to Data" (RS line high).
;       Enter at "Strobe" for a simple pulse.


WriteData       andlw   B'11110000'     ; set control lines to zero, and
                xorlw   B'00000001'     ; set RS line high for "Write to DD"
Strobe          movwf   PORTB
                bsf     PORTB,1         ; raise Enable bit
                nop
                nop
                bcf     PORTB,1         ; and drop it again
                movlw   .22             ; set up for 5 mS delay
                call    Delay
                return

; -------------------------------------------------------------------
; Blank the LCD

Blank           movlw   B'00000000'     ; Set control lines to zero
                call    Strobe          ; send high nibble
                movlw   B'00010000'     ; Set control lines to zero
                call    Strobe
                return


;-------------------------------------------------------
; Delay routine: load multiples of 244uS in W

Delay           movwf   High_Count
Del01           movlw   .122            ; Clock is 1 Mhz, gives 1 uS step
                movwf   Low_Count       ; this loop takes 2us x 122 = 244 uS
Del02           decfsz  Low_Count,F
                goto    Del02
                decfsz  High_Count,F
                goto    Del01
                return

; ---------------------------------------------------------
; Display five decimal digits packed into MSD, MID, LSD

Disp_Big        movf    MSD,W                   ; get MSD
                call    Ascii

                swapf   MID,W                   ; move high byte down
                andlw   B'00001111'             ; mask out low nybble
                call    Ascii
       
                movf    MID,W                   ; get low byte
                andlw   B'00001111'             ; mask out low nybble
                call    Ascii

                swapf   LSD,W                   ; move high byte down
                andlw   B'00001111'             ; mask out low nybble
                call    Ascii
       
                movf    LSD,W                   ; get low byte
                andlw   B'00001111'             ; mask out low nybble
                call    Ascii

                return

;------------------------------------------------------------
;      Convert the 8 bit binary number in W Register
;      to a 3 digit BCD number stored in LSB, MID & MSB.

BinBCD          movwf   LSD
                clrf    MSD             ; Clear all output regs
                clrf    MID

tens            movlw   .10
                subwf   LSD,W           ; Subtract ten from amount
                BTFSS   STATUS,C
                goto    hunds
                movwf   LSD             ; save remainder
                incf    MID,F           ; Inc Tens
                goto    tens            ; Until finished

hunds           movlw   .10
                subwf   MID,W           ; Subtract ten from amount
                BTFSS   STATUS,C
                return
                movwf   MID             ; save remainder
                incf    MSD,F           ; Inc Tens
                goto    hunds           ; Until finished

finn            return

