;   Temp.ASM
;             PIC Temperature and Humidity measurement.
;
;      Use a Dallas DS1620 IC to measure the temperature,
;             a HS15P sensor to measure humidity
;         and display the results on the LCD display
;
; Pins used:
;               DS1620 Temp IC
;               RA0 = Data i/o
;               RA1 = Clock line (normally high)
;               RA2 = Reset line (normally low)
;
;               HS15P Humid Sensor
;               RB2 & Rb3 are drive lines to Sensor
;               RA3 is input to A/D for resistance measurement
;
; ---------------------------------------------------------------------------
; Measure Temperature and Humidity (in Deg C).

MeasureTempC    call    SetupTemp       ; init everything
TempC01         call    ReadTempC       ; get temperature data in C
                call    HumidRead       ; get humidity data
                goto    TempC01

; ---------------------------------------------------------------------------
; Measure Temperature and Humidity (in Deg F).

MeasureTempF    call    SetupTemp       ; init everything
TempF01         call    Fahrenheit      ; get temperature data in F
                call    HumidRead       ; get humidity data
                goto    TempF01

; ---------------------------------------------------------------------------
; Fahrenheit conversion:

; The DS1620 IC reads temperature in Degrees Centigrade from -55C to +125C
; in 1/2 deg steps. It supplies the data in binary "two's compliment" form.
; This form would be ideal for Fahrenheit conversion using Floating Point maths,
; however with this project we have run out of room to include a FP package
; and instead will have to make a few compromises. We will do the conversion
; using a "look-up" table, and will need to restrict the temp range from
; 0->75 deg C (approx 32->167 deg F). We will not include the "decimal point".

        org     H'200'                  ; put the tables after a page boundary
                                        ; to simplify "Table Reads"

Fahrenheit      movlw   HIGH FahTable   ; Initialise PCLATH
                movwf   PCLATH          ; for table read

                call    SerInOut        ; get data
                movlw   H'07'           ; point to LCD
                call    Position
                btfss   Temp2,0         ; was reading negative (eg below zero dec C)?
                goto    ReadF01         ; no
                clrf    TempComp        ; yes, set temp compensation to zero
                clrf    Temp1           ; set temp to zero

ReadF01         call    TempLine        ; select temperature compensation (for humoidity)
                bcf     STATUS,C
                rrf     Temp1,F         ; shift "1/2 deg" bit out
                movf    Temp1,W
                sublw   D'75'           ; is it greater than 75 deg C ?
                btfsc   STATUS,C
                goto    ReadF03         ; no.
                movlw   D'75'           ; yes, limit it to 75
                movwf   Temp1

ReadF03         call    FahTable        ; convert from C to F
                call    BinBCD          ; convert to BCD
                call    No_Zeros        ; display with no leading zero
                call    PrintSpace
                movlw   A'F'            ; "degrees F"
                call    Display
                call    PrintSpace      ; wipe out old trailing chars
                call    PrintSpace
                return

; ---------------------------------------------------------------------------
FahTable        movf    Temp1,W         ; get temperature
                addwf   PCL,1           ; return deg F, given deg C
                
                                    ; deg C         deg F
          retlw     D'32'           ; 0             32
          retlw     D'34'           ; 1             33.8
          retlw     D'36'           ; 2             35.6
          retlw     D'37'           ; 3             37.4
          retlw     D'39'           ; 4             39.2
          retlw     D'41'           ; 5             41
          retlw     D'43'           ; 6             42.8
          retlw     D'45'           ; 7             44.6
          retlw     D'46'           ; 8             46.4
          retlw     D'48'           ; 9             48.2
          retlw     D'50'           ; 10            50
          retlw     D'52'           ; 11            51.8
          retlw     D'54'           ; 12            53.6
          retlw     D'55'           ; 13            55.4
          retlw     D'57'           ; 14            57.2
          retlw     D'59'           ; 15            59
          retlw     D'61'           ; 16            60.8
          retlw     D'63'           ; 17            62.6
          retlw     D'64'           ; 18            64.4
          retlw     D'66'           ; 19            66.2
          retlw     D'68'           ; 20            68
          retlw     D'70'           ; 21            69.8
          retlw     D'72'           ; 22            71.6
          retlw     D'73'           ; 23            73.4
          retlw     D'75'           ; 24            75.2
          retlw     D'77'           ; 25            77
          retlw     D'79'           ; 26            78.8
          retlw     D'81'           ; 27            80.6
          retlw     D'82'           ; 28            82.4
          retlw     D'84'           ; 29            84.2
          retlw     D'86'           ; 30            86
          retlw     D'88'           ; 31            87.8
          retlw     D'90'           ; 32            89.6
          retlw     D'91'           ; 33            91.4
          retlw     D'93'           ; 34            93.2
          retlw     D'95'           ; 35            95
          retlw     D'97'           ; 36            96.8
          retlw     D'99'           ; 37            98.6
          retlw     D'100'          ; 38            100.4
          retlw     D'102'          ; 39            102.2
          retlw     D'104'          ; 40            104
          retlw     D'106'          ; 41            105.8
          retlw     D'108'          ; 42            107.6
          retlw     D'109'          ; 43            109.4
          retlw     D'111'          ; 44            111.2
          retlw     D'113'          ; 45            113
          retlw     D'115'          ; 46            114.8
          retlw     D'117'          ; 47            116.6
          retlw     D'118'          ; 48            118.4
          retlw     D'120'          ; 49            120.2
          retlw     D'122'          ; 50            122
          retlw     D'124'          ; 51            123.8
          retlw     D'126'          ; 52            125.6
          retlw     D'127'          ; 53            127.4
          retlw     D'129'          ; 54            129.2
          retlw     D'131'          ; 55            131
          retlw     D'133'          ; 56            132.8
          retlw     D'135'          ; 57            134.6
          retlw     D'136'          ; 58            136.4
          retlw     D'138'          ; 59            138.2
          retlw     D'140'          ; 60            140
          retlw     D'142'          ; 61            141.8
          retlw     D'144'          ; 62            143.6
          retlw     D'145'          ; 63            145.4
          retlw     D'147'          ; 64            147.2
          retlw     D'149'          ; 65            149
          retlw     D'151'          ; 66            150.8
          retlw     D'153'          ; 67            152.6
          retlw     D'154'          ; 68            154.4
          retlw     D'156'          ; 69            156.2
          retlw     D'158'          ; 70            158
          retlw     D'160'          ; 71            159.8
          retlw     D'162'          ; 72            161.6
          retlw     D'163'          ; 73            163.4
          retlw     D'165'          ; 74            165.2
          retlw     D'167'          ; 75            167

; ---------------------------------------------------------------------------
; Read and display temp in deg C

ReadTempC       call    SerInOut        ; get data
                movlw   H'07'           ; point to LCD
                call    Position
                btfss   Temp2,0         ; was reading negative (eg below zero dec C)?
                goto    Read001         ; no

                movlw   H'2D'           ; yes, so print "-"
                call    Display
                clrf    TempComp        ; set temp compensation to zero (for humidity)
                bsf     STATUS,C        ; keep leading one (two's comp)
                comf    Temp1,F         ; convert from two's compliment form
                incf    Temp1,F
                bcf     STATUS,C
                rrf     Temp1,W         ; shift "1/2 deg" bit out
                call    BinBCD          ; convert to BCD
                goto    Read003

Read001         movlw   H'2B'           ; it's positive, so print "+"
                call    Display
                call    TempLine        ; calc temp compensation (for Humidity)

Read003         call    No_Zeros        ; display with no leading zero
                movlw   D'46'           ; "."
                call    Display
                btfss   Temp1,0         ; is "1/2 deg" bit set?
                goto    Read004
                movlw   A'5'
                goto    Read005
Read004         movlw   A'0'            ; "0"
Read005         call    Display
                call    PrintSpace
                movlw   A'C'            ; "degrees C"
                call    Display
                call    PrintSpace      ; wipe out old trailing chars
                call    PrintSpace
                call    PrintSpace
                return

; ---------------------------------------------------------------------------
; Save the "Tens of degree C" value to use for Humidity calculation.

TempLine        bcf     STATUS,C
                rrf     Temp1,W         ; shift "1/2 deg" bit out
                call    BinBCD          ; convert to BCD
                movf    MID,W
                movwf   TempComp        ; save the "tens of deg C" value
                movf    MSD,W
                btfsc   STATUS,Z        ; if it's over 100 deg C
                return
                movlw   D'09'           ; use "90 deg C" for compensation
                movwf   TempComp
                return

; ---------------------------------------------------------------------------
; A routine to calculate the humidity, given the A/D reading (0->255).
; The humidity sensor data sheet gives curves for Humid v/s Resistance,
; at 5 different temperatures. We have re-plotted these against A/D readings,
; and then extracted the data for 5 look-up tables.
; A first table lists the "break-points" for the A/D resistance readings
; and returns a pointer between 0 and 14.
; This then points to the actual humidity values in the 5 Humidity tables.
; The measured temperature is used to decide which Humidity table is used.

                org     H'300'          ; put the tables after a page boundary
                                        ; to simplify "Table Reads"

HumidCalc       movlw   HIGH CountTable ; Initialise PCLATH
                movwf   PCLATH          ; for table read

                movlw   H'00'           ; initialise Humidity pointer
                movwf   Pointer

; convert Resistance to a table pointer value

HCalc001        movf    Pointer,W       ; get current pointer
                call    CountTable      ; get a reading
                subwf   Humidity,W      ; is this value equal (or greater?)
                btfsc   STATUS,C
                goto    HCalc002        ; yes
                incf    Pointer,F       ; not yet, try next value
                goto    HCalc001

; decide which temperature/table to use

HCalc002        movf    TempComp,W      ; is it 0->10 deg C ? (0)
                btfsc   STATUS,Z
                goto    HCalc003        ; yes.
                sublw   D'01'           ; is it 10->20 deg C ? (1)
                btfsc   STATUS,Z
                goto    HCalc004        ; yes.
                sublw   D'02'           ; is it 20->30 deg C ? (2)
                btfsc   STATUS,Z
                goto    HCalc005        ; yes.
                sublw   D'03'           ; is it 30->40 deg C ? (3)
                btfsc   STATUS,Z
                goto    HCalc006        ; yes.
                goto    HCalc007        ; it must be over 40 deg C (4->9)

; look up humidity in the table

HCalc003        call    T5_Table        ; get 5 deg Humidity
                goto    HCalc008
HCalc004        call    T15_Table       ; get 15 deg Humidity
                goto    HCalc008
HCalc005        call    T25_Table       ; get 25 deg Humidity
                goto    HCalc008
HCalc006        call    T35_Table       ; get 35 deg Humidity
                goto    HCalc008
HCalc007        call    T45_Table       ; get 45 deg Humidity
HCalc008        movwf   Humidity        ; save result
                return

; ---------------------------------------------------------------------------
; Read AN556 in Embedded Control Handbook to understand "Table Reads".

CountTable      movf    Pointer,W
                addwf   PCL,1           ; return a pointer, given an A/D reading
                retlw  D'255'           ; 0     (pointer)
                retlw  D'254'           ; 1
                retlw  D'252'           ; 2
                retlw  D'249'           ; 3
                retlw  D'242'           ; 4
                retlw  D'225'           ; 5
                retlw  D'201'           ; 6
                retlw  D'166'           ; 7
                retlw  D'108'           ; 8
                retlw  D'69'            ; 9
                retlw  D'40'            ; 10
                retlw  D'17'            ; 11
                retlw  D'9'             ; 12
                retlw  D'4'             ; 13
                retlw  D'0'             ; 14
                
T5_Table        movf    Pointer,W
                addwf  PCL,1            ; return Humidity, given a pointer
                retlw  D'30'            ; 0  (pointer)
                retlw  D'35'            ; 1
                retlw  D'40'            ; 2
                retlw  D'45'            ; 3
                retlw  D'50'            ; 4
                retlw  D'60'            ; 5
                retlw  D'65'            ; 6
                retlw  D'70'            ; 7
                retlw  D'75'            ; 8
                retlw  D'90'            ; 9
                retlw  D'95'            ; 10
                retlw  D'100'           ; 11
                retlw  D'100'           ; 12
                retlw  D'100'           ; 13
                retlw  D'100'           ; 14

T15_Table       movf    Pointer,W
                addwf  PCL,1            ; return Humidity, given a pointer
                retlw  D'25'            ; 0  (pointer)
                retlw  D'30'            ; 1
                retlw  D'35'            ; 2
                retlw  D'40'            ; 3
                retlw  D'45'            ; 4
                retlw  D'50'            ; 5
                retlw  D'55'            ; 6
                retlw  D'65'            ; 7
                retlw  D'75'            ; 8
                retlw  D'80'            ; 9
                retlw  D'90'            ; 10
                retlw  D'100'           ; 11
                retlw  D'100'           ; 12
                retlw  D'100'           ; 13
                retlw  D'100'           ; 14

T25_Table       movf    Pointer,W
                addwf  PCL,1            ; return Humidity, given a pointer
                retlw  D'20'            ; 0  (pointer)
                retlw  D'25'            ; 1
                retlw  D'30'            ; 2
                retlw  D'30'            ; 3
                retlw  D'35'            ; 4
                retlw  D'40'            ; 5
                retlw  D'50'            ; 6
                retlw  D'55'            ; 7
                retlw  D'65'            ; 8
                retlw  D'70'            ; 9
                retlw  D'80'            ; 10
                retlw  D'90'            ; 11
                retlw  D'95'            ; 12
                retlw  D'100'           ; 13
                retlw  D'100'           ; 14

T35_Table       movf    Pointer,W
                addwf  PCL,1            ; return Humidity, given a pointer
                retlw  D'15'            ; 0  (pointer)
                retlw  D'20'            ; 1
                retlw  D'20'            ; 2
                retlw  D'25'            ; 3
                retlw  D'30'            ; 4
                retlw  D'35'            ; 5
                retlw  D'40'            ; 6
                retlw  D'50'            ; 7
                retlw  D'55'            ; 8
                retlw  D'65'            ; 9
                retlw  D'75'            ; 10
                retlw  D'85'            ; 11
                retlw  D'95'            ; 12
                retlw  D'100'           ; 13
                retlw  D'100'           ; 14

T45_Table       movf    Pointer,W
                addwf  PCL,1            ; return Humidity, given a pointer
                retlw  D'5'             ; 0  (pointer)
                retlw  D'10'            ; 1
                retlw  D'15'            ; 2
                retlw  D'20'            ; 3
                retlw  D'25'            ; 4
                retlw  D'30'            ; 5
                retlw  D'35'            ; 6
                retlw  D'40'            ; 7
                retlw  D'50'            ; 8
                retlw  D'55'            ; 9
                retlw  D'65'            ; 10
                retlw  D'75'            ; 11
                retlw  D'85'            ; 12
                retlw  D'95'            ; 13
                retlw  D'100'           ; 14

; ---------------------------------------------------------------------------
; Initialise PIC for Temp/Humidity reading:

SetupTemp       bsf     STATUS,RP0      ; Select page 1
                movlw   B'00000011'     ; select digital I/O
                movwf   ADCON0          ; (actually ADCON1 reg in page 1)
                movlw   B'00001000'     ; data direction Port A
                movwf   PORTA
                movlw   B'00000000'     ; data direction Port B
                movwf   PORTB
                bcf     STATUS,RP0      ; back to page 0 of RAM
                movlw   B'11011001'     ; Select input from Ra3
                movwf   ADCON0

                movlw   B'00000010'     ; set control lines for DS1620
                movwf   PORTA           ; (clock high, Data & Reset low)

                call    Temp_S          ; print "Temp = "
                call    Humid_S         ; print "Humid = "

; initialize the DS1620

                bsf     PORTA,2         ; set Reset line high to begin communications
                movlw   H'0C'           ; send "Write Configuration byte"
                call    SerOut
                movlw   B'00000010'     ; send "set up for continuous conversions"
                call    SerOut
                bcf     PORTA,2         ; set Reset line low
                nop
                nop
                bsf     PORTA,2         ; set Reset line high
                movlw   H'EE'           ; send "Start Conversions"
                call    SerOut
                bcf     PORTA,2         ; set Reset line low
                return

; ---------------------------------------------------------------------------
; Serial Output routine for DS1620

SerOut          movwf   Temp1           ; save the data
                movlw   D'08'           ; set up for 8 bits transferred
                movwf   Temp2

Ser001          bcf     PORTA,1         ; drop clock line
                btfss   Temp1,0         ; test next bit
                goto    Ser002
                bsf     PORTA,0         ; set data high
                goto    Ser003
Ser002          bcf     PORTA,0         ; set data low
Ser003          bsf     PORTA,1         ; set clock high
                rrf     Temp1,F         ; shift next bit across
                decfsz  Temp2,F         ; have we finished 8 bits?
                goto    Ser001          ; no, go around again
                return
                      
; ---------------------------------------------------------------------------
; Serial Input/Output routine for 1620

SerInOut        movlw   H'AA'           ; send "Read Temperature"
                movwf   Temp1
                movlw   B'00001000'     ; switch RA0 to output
                tris    PORTA           ; (leave RA4 as input)
                movlw   D'08'           ; set up for 8 bits transferred
                movwf   Temp2
                bsf     PORTA,2         ; set Reset line high to begin communications

; send the "Read" command

SerI001         bcf     PORTA,1         ; drop clock line
                bsf     PORTA,0         ; assume data bit high
                btfss   Temp1,0         ; test next bit
                bcf     PORTA,0         ; set data line low
                bsf     PORTA,1         ; set clock line high
                rrf     Temp1,F         ; shift next bit across
                decfsz  Temp2,F         ; have we finished 8 bits?
                goto    SerI001         ; no, go around again

; get first byte

                movlw   B'00001001'     ; switch RA0 back to input
                tris    PORTA           ; (leave RA4 as input)
                movlw   D'08'           ; set up for 8 bits transferred
                movwf   Temp2

SerI005         rrf     Temp1,F         ; shift previous bit across
                bcf     PORTA,1         ; drop clock line
                bcf     Temp1,7         ; assume bit is low
                btfsc   PORTA,0         ; test incoming bit
                bsf     Temp1,7         ; it was high
                bsf     PORTA,1         ; set clock high
                decfsz  Temp2,F         ; have we finished 8 bits?
                goto    SerI005         ; no, go around again

; get 9th bit
                bcf     PORTA,1         ; drop clock line
                clrf    Temp2
                btfsc   PORTA,0         ; test next bit
                bsf     Temp2,0         ; it was high
                bsf     PORTA,1         ; set clock high
                bcf     PORTA,2         ; drop Reset line
                return
                      
; ---------------------------------------------------------------------------
; Humidity measurement.
;
; The HS15P humidity sensor responds to Humidity changes by varying its
; Resistance. It is essential that the sensor does not have a standing DC
; voltage across it as the resulting electrolysis will quickly destroy the
; device. We overcome this by doing a dummy measurement in one direction,
; then reading again in the reverse direction, and then disconnecting it.
; This ensures that the sensor has only AC applied.

HumidRead       bsf     STATUS,RP0              ; Select page 1
                movlw   B'00000000'             ; select all analog inputs
                movwf   ADCON0                  ; (actually is ADCON1 reg in page 1)
                bcf     STATUS,RP0              ; back to page 0

; first do a dummy read

                movlw   B'00000100'             ; set RB2 low, RB3 high
                movwf   PORTB                   
                movlw   H'10'                   ; wait till volts settle down
                call    Delay
                call    Volt                    ; get reading from Humidity sensor
                movwf   Temp1

; then reverse the connections

                movlw   B'00001000'             ; set RB2 high, RB3 low
                movwf   PORTB
                movlw   H'10'                   ; wait till volts settle down
                call    Delay
                call    Volt                    ; get reading from potentiometer
                movwf   Humidity
                movlw   B'00000000'             ; set both lines low again
                movwf   PORTB

; now disconnect it

                bsf     STATUS,RP0              ; Select page 1
                movlw   B'00000011'             ; put inputs back to digital
                movwf   ADCON0
                bcf     STATUS,RP0              ; back to page 0

                call    HumidCalc               ; get Humidity value from tables
                movlw   D'72'                   ; point to LCD
                call    Position
                movf    Humidity,W
                call    BinBCD                  ; convert to decimal
                call    No_Zeros                ; and display result
                call    PrintSpace
                movlw   H'25'                   ; "%"
                call    Display
                call    PrintSpace              ; wipe out old trailing chars
                call    PrintSpace
                return

; ---------------------------------------------------------------------------


